#include "sqi.h"

/********************************************
 * _sqiOpenDatabase
 *
 * This func does two things...
 * 1. loads the Database Definition and
 * 2. makes calls to load table data
 *
 * Database Definition
 * -------------------
 * This file is the database to the user. It can
 * also be referred to as the catalog file. Here
 * is a sample to show what you may find inside...
 *
 
[DATABASE]
name=contacts
description=all tables needed for a contacts database
defaultpath=/home/pharvey/projects/sqi/test

[TABLE]
name=contacts
description=list of all contacts
fixedcolumns=no
delimiter=|
file=/home/pharvey/projects/sqi/test/contacts.tb

[COLUMN]
name=pkContact
description=
type=numeric
size=4

[COLUMN]
name=FirstName
description=
type=varchar
size=50

[COLUMN]
name=LastName
description=
type=varchar
size=50
 
[TABLE]
name=logs
description=activity log
fixedcolumns=no
delimiter=|
file=logs.tb

[COLUMN]
name=pkEntry
description=
type=date
size=

[COLUMN]
name=activity
description=whatever was done
type=varchar
size=500
 
 * And here are some rules for this file...
 * 1. stuff between [] are OBJECTS
 * 2. LValue= are PROPERTIES
 * 3. RValue are PROPERTY VALUES
 * 4. Objects are case insensitive
 * 5. Properties are case insensitive
 * 6. Property Values are case insensitive where possible.
 *
 * Here are some rules for Objects...
 * 1. The database object must be first
 * 2. There can only be one database object.
 * 3. Table objects must be followed by their columns
 * 4. There can be more than one table object.
 * 5. Column order is important (must reflect the order of the data in TABLE.file)
 *
 * Also...
 * 1. Table names should be unique within database.
 * 2. Column names should be unique within table.
 ********************************************/

int _sqiGetDatabaseProperties( HSQI hSQI, HINI hIni );
int _sqiGetTableProperties( HSQI hSQI, HINI hIni, HSQITABLE *phTable );
int _sqiGetColumnProperties( HSQI hSQI, HINI hIni, HSQITABLE hTable );


int _sqiOpenDatabase( HSQI hSQI )
{
	HINI		hIni;
	char		szObject[INI_MAX_OBJECT_NAME+1];
	int			bCreate	= false;
	HSQITABLE	hTable	= 0;
    FILE    	*hFile;

	hSQI->hTableCache	= lstOpen();
	lstSetFreeFunc( hSQI->hTableCache, _sqiCloseTable );

	if ( iniOpen( &hIni, hSQI->pszCatalogFile, ';', '[', ']', '=', bCreate ) != INI_SUCCESS )
	{
		sprintf( hSQI->szMsg, "Unable to open database file %s", hSQI->pszCatalogFile );
		logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, hSQI->szMsg );
		return false;
	}
	
	/* DATABASE OBJECT MUST BE AT TOP SO... */
	iniObjectFirst( hIni );
	if ( iniObjectEOL( hIni ) )
	{
		sprintf( hSQI->szMsg, "(%s) does not appear to be in INI format or is empty.", hSQI->pszCatalogFile );
		logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, hSQI->szMsg );
		return false;
	}
	
	*szObject = '\0';
	iniObject( hIni, szObject );
	if ( strcasecmp( szObject, "DATABASE" ) != 0 )
	{
		sprintf( hSQI->szMsg, "[DATABASE] object not found at top of %s. Perhaps not a database file.", hSQI->pszCatalogFile );
		logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, hSQI->szMsg );
		return false;
	}

	if ( !_sqiGetDatabaseProperties( hSQI, hIni ) )
	{
		logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "Unable to read database properties." );
		return false;
	}


	/* NOW GET ALL OTHER OBJECTS */
	iniObjectNext( hIni );
	while ( !iniObjectEOL( hIni ) )
	{
		*szObject = '\0';
		iniObject( hIni, szObject );

		if ( strcasecmp( szObject, "TABLE" ) == 0 )
		{
			hTable = 0;
			if ( _sqiGetTableProperties( hSQI, hIni, &hTable ) )
				lstAppend( hSQI->hTableCache, hTable );
			else
				logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "Failed to get table properties." );
		}
		else if ( strcasecmp( szObject, "COLUMN" ) == 0 )
		{
			if ( !hTable )
				logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "Invalid column position. Column definition must follow its table." );
			else if ( !_sqiGetColumnProperties( hSQI, hIni, hTable ) )
				logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "Failed to get column properties." );
		}
		else
		{
			hTable = 0;
			sprintf( hSQI->szMsg, "Invalid object %s", szObject );
			logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, hSQI->szMsg );
		}

		iniObjectNext( hIni );
	}
	
	/* LETS PRELOAD ALL TABLE DATA */
	lstFirst( hSQI->hTableCache );
	while ( !lstEOL( hSQI->hTableCache ) )
	{
		hTable	= (HSQITABLE)lstGet( hSQI->hTableCache );
		hFile	= 0;
		if ( hTable->pszFileName )
		{
			hFile = fopen( hTable->pszFileName, "r" );
			if ( !hFile )
				hFile = fopen( hTable->pszFileName, "w" );
	
			if ( hFile )
				_sqiLoadTableData( hTable, hFile );
			else	
			{
				sprintf( hSQI->szMsg, "Unable to open %s", hTable->pszFileName );
				logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, hSQI->szMsg );
			}
	
			fclose( hFile );
		}

		lstNext( hSQI->hTableCache );
	}
	
	
	iniClose( hIni );

	return true;
}


int _sqiGetDatabaseProperties( HSQI hSQI, HINI hIni )
{
	char		szProperty[INI_MAX_PROPERTY_NAME+1];
	char		szPropertyValue[INI_MAX_PROPERTY_VALUE+1];

	iniPropertyFirst( hIni );
	while ( !iniPropertyEOL( hIni ) )
	{
		*szProperty			= '\0';
		*szPropertyValue	= '\0';
		iniProperty( hIni, szProperty );

		if ( strcasecmp( szProperty, "NAME" ) == 0 )
		{
			if ( hSQI->pszName )
				logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "NAME property repeats in database object." );
			else
			{
				iniValue( hIni, szPropertyValue );
				hSQI->pszName = (char*)strdup( szPropertyValue );
			}
		}
		else if ( strcasecmp( szProperty, "DESCRIPTION" ) == 0 )
		{
			if ( hSQI->pszDescription )
				logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "DESCRIPTION property repeats in database object." );
			else
			{
				iniValue( hIni, szPropertyValue );
				hSQI->pszDescription = (char*)strdup( szPropertyValue );
			}
		}
		else if ( strcasecmp( szProperty, "DEFAULTPATH" ) == 0 )
		{
			if ( hSQI->pszDefaultPath )
				logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "DEFAULTPATH property repeats in database object." );
			else
			{
				iniValue( hIni, szPropertyValue );
				hSQI->pszDefaultPath = (char*)strdup( szPropertyValue );
			}
		}
		else
		{
			sprintf( hSQI->szMsg, "(%s) not a valid property for database object.", szProperty );
			logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, hSQI->szMsg );
		}

		iniPropertyNext( hIni );
	}
	
	/******************************
	 * DO SOME VALIDATION
	 ******************************/
	if ( !(hSQI->pszName) || !(hSQI->pszName[0]) )
		logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "Database object is missing the name property." );

	return true;
}

int _sqiGetTableProperties( HSQI hSQI, HINI hIni, HSQITABLE *phTable )
{
	HSQITABLE	hTable;
	char		szProperty[INI_MAX_PROPERTY_NAME+1];
	char		szPropertyValue[INI_MAX_PROPERTY_VALUE+1];

	/******************************
	 * INIT NEW HSQITABLE
	 ******************************/
	hTable = malloc( sizeof(SQITABLE) );
	hTable->pszFileName 		= 0;
	hTable->pszTableName		= 0;
	hTable->pszDescription		= 0;
	hTable->bFixedColumns 		= 0;
	hTable->cDelimitChar 		= '|';
	hTable->hColHdrs			= lstOpen();
	hTable->hRows				= lstOpen();
	hTable->nRefs				= 1;				/* init to 1 if you want to keep the table loaded even when no refs... ie for a table cache */
	hTable->hBase				= NULL;

	/******************************
	 * USE CUSTOM FREE FUNCS TO ENSURE ALL MEM IS FREED
	 ******************************/
	lstSetFreeFunc( hTable->hColHdrs, _sqiFreeColumnHeader );
	lstSetFreeFunc( hTable->hRows, _sqiFreeDataRow );
	
	/******************************
	 * GET VALUES
	 ******************************/
	iniPropertyFirst( hIni );
	while ( !iniPropertyEOL( hIni ) )
	{
		*szProperty			= '\0';
		*szPropertyValue	= '\0';
		iniProperty( hIni, szProperty );

		if ( strcasecmp( szProperty, "NAME" ) == 0 )
		{
			if ( hTable->pszTableName )
				logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "NAME property repeats in table object." );
			else
			{
				iniValue( hIni, szPropertyValue );
				hTable->pszTableName = (char*)strdup( szPropertyValue );
			}
		}
		else if ( strcasecmp( szProperty, "DESCRIPTION" ) == 0 )
		{
			if ( hTable->pszDescription )
				logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "DESCRIPTION property repeats in table object." );
			else
			{
				iniValue( hIni, szPropertyValue );
				hTable->pszDescription = (char*)strdup( szPropertyValue );
			}
		}
		else if ( strcasecmp( szProperty, "FIXEDCOLUMNS" ) == 0 )
		{
			iniValue( hIni, szPropertyValue );
			if ( szPropertyValue[0] == '1' || toupper(szPropertyValue[0]) == 'Y' || toupper(szPropertyValue[0]) == 'T' )
				hTable->bFixedColumns = 1;
		}
		else if ( strcasecmp( szProperty, "DELIMITER" ) == 0 )
		{
			iniValue( hIni, szPropertyValue );
			hTable->cDelimitChar = szPropertyValue[0];
		}
		else if ( strcasecmp( szProperty, "FILE" ) == 0 )
		{
			if ( hTable->pszFileName )
				logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "FILE property repeats in table object." );
			else
			{
				iniValue( hIni, szPropertyValue );
				hTable->pszFileName = (char*)strdup( szPropertyValue );
			}
		}
		else
		{
			sprintf( hSQI->szMsg, "(%s) not a valid property for table object.", szProperty );
			logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, hSQI->szMsg );
		}

		iniPropertyNext( hIni );
	}
	
	/******************************
	 * DO SOME VALIDATION
	 ******************************/
	if ( !(hTable->pszTableName) || !(hTable->pszTableName[0]) )
		logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "Table object is missing the NAME property." );
	else if ( !(hTable->pszFileName) || !(hTable->pszFileName[0]) )
		logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "Table object is missing the FILE property." );
	else if ( !(hTable->cDelimitChar) && !(hTable->bFixedColumns) )
	{
        hTable->bFixedColumns = 1;
		logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "FixedColumns or DelimitChar must be specified. Trying FixedColumns." );
	}

    *phTable = hTable;

	return true;
}

int _sqiGetColumnProperties( HSQI hSQI, HINI hIni, HSQITABLE hTable )
{
	char		szProperty[INI_MAX_PROPERTY_NAME+1];
	char		szName[INI_MAX_PROPERTY_VALUE+1]			= "";
	char		szDescription[INI_MAX_PROPERTY_VALUE+1]		= "";
	char		szType[INI_MAX_PROPERTY_VALUE+1]			= "";
	char		szSize[INI_MAX_PROPERTY_VALUE+1]			= "";
	int			nSize										= 255;

	/******************************
	 * GET VALUES
	 ******************************/
	iniPropertyFirst( hIni );
	while ( !iniPropertyEOL( hIni ) )
	{
		*szProperty			= '\0';
		iniProperty( hIni, szProperty );

		if ( strcasecmp( szProperty, "NAME" ) == 0 )
		{
			if ( szName[0] )
				logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "NAME property repeats in column object." );
			else
				iniValue( hIni, szName );
		}
		else if ( strcasecmp( szProperty, "DESCRIPTION" ) == 0 )
		{
			if ( szDescription[0] )
				logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "DESCRIPTION property repeats in column object." );
			else
				iniValue( hIni, szDescription );
		}
		else if ( strcasecmp( szProperty, "TYPE" ) == 0 )
		{
			if ( szType[0] )
				logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "TYPE property repeats in column object." );
			else
				iniValue( hIni, szType );
		}
		else if ( strcasecmp( szProperty, "SIZE" ) == 0 )
		{
			if ( szSize[0] )
				logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "SIZE property repeats in column object." );
			else
			{
				iniValue( hIni, szSize );
				nSize = strtod( szSize, NULL );
				if ( nSize < 1 || nSize > 32000 )
					nSize = 255;
			}
		}
		else
		{
			sprintf( hSQI->szMsg, "(%s) not a valid property for column object.", szProperty );
			logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, hSQI->szMsg );
		}

		iniPropertyNext( hIni );
	}
	
	/******************************
	 * DO SOME VALIDATION
	 ******************************/
	if ( !(szName[0]) )
		logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "Column object is missing the NAME property." );
	else if ( !(szType[0]) )
		logPushMsg( hSQI->hLog, __FILE__, __FILE__, __LINE__, LOG_INFO, SQI_ERROR, "Column object is missing the TYPE property." );

	lstAppend( hTable->hColHdrs, _sqiNewColumnHeader( hTable->hColHdrs->nItems, hSQI->pszName, hTable->pszTableName, szName, SQL_VARCHAR, nSize ) );

	return true;
}

